/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import unikl.disco.misc.MapValueComparator;
import unikl.disco.misc.SetUtils;

import edu.uci.ics.jung.graph.DirectedEdge;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.impl.DirectedSparseEdge;
import edu.uci.ics.jung.graph.impl.DirectedSparseGraph;
import edu.uci.ics.jung.graph.impl.SimpleSparseVertex;
import edu.uci.ics.jung.utils.Pair;

/**
 * Class implementing the full version of the turn-prohibition algorithm described in
 * <ul>
 * <li>D. Starobinsky, M. Karpovsky, and L. Zakrevski: <it>Application of Network Calculus
 * to General Topologies using Turn-Prohibition</it>.
 * </ul>
 * Contains methods for determining the sets of prohibited (and permitted) turns and to
 * update a given server graph accordingly.
 * 
 * @author Frank A. Zdarsky
 */
public class TurnProhibition {
	/** The network graph on which the turn prohibition algorithm runs */
	protected DirectedSparseGraph network_graph;
	/** A comparator comparing the degree of nodes */
	protected Comparator node_comparator;
	/** The set of nodes visited by the algorithm */
	protected Set visited_nodes;
	/** The set of links visited by the algorithm */
	protected Set visited_links;
	/** The set of permitted turns */
	protected Set permitted_turns;
	/** The set of prohibited turns */
	protected Set prohibited_turns;
	/** A map of nodes to node degrees */
	protected Map node_degrees;
	/** A queue of nodes sorted by ascending node degree */
	protected List node_queue;

	/**
	 * Creates a turn prohibition instance for the specified network graph
	 * which must be bidirectional (i.e. a directional in which for each link
	 * (u,v) there exists exactly one link (v,u)). 
	 * @param network_graph
	 */
    public TurnProhibition(DirectedSparseGraph network_graph) {
    	this.network_graph = network_graph;
    }

    /**
     * Initializes this instance by resetting all member variables and
     * creating the node queue and the node-to-node-degree-map.
     */
    private void initialize() {
    	visited_nodes    = new HashSet();
    	visited_links    = new HashSet();
    	permitted_turns  = new HashSet();
    	prohibited_turns = new HashSet();
    	node_degrees     = new HashMap();
    	node_queue       = new ArrayList();
    	node_comparator  = new MapValueComparator(node_degrees);
    	for (Iterator iter = network_graph.getVertices().iterator(); iter.hasNext(); ) {
    		Vertex v = (Vertex) iter.next();
    		node_degrees.put(v, new Integer(v.getNeighbors().size()));
    		node_queue.add(v);
    	}
    	Collections.sort(node_queue, node_comparator);
    }
    
    /**
     * Applies the turn-prohibition algorithm to the current network graph and
     * returns a set of prohibited turns. Turns are stored as instances of
     * <code>edu.uci.ics.jung.utils.Pair</code> containing the two edges of the
     * network graph that constitute the turn.
     * @return a set of prohibited turns
     */
    public Set runTurnProhibition() {
    	initialize();    	
    	runTurnProhibitionInternal(network_graph.getVertices(), null);
    	return prohibited_turns;
    }
    
    /**
     * Returns the set of prohibited turns (turn = a <code>Pair</code> instance containing
     * two edges) computed during the last invokation of <code>runTurnProhibition()</code>.
     * @return a set of prohibited turns
     */
    public Set getProhibitedTurns() {
    	return prohibited_turns;
    }
    
    /**
     * Returns the set of permitted turns (turn = a <code>Pair</code> instance containing
     * two edges) computed during the last invokation of <code>runTurnProhibition()</code>.
     * @return a set of permitted turns
     */
    public Set getPermittedTurns() {
    	return permitted_turns;
    }
    
    /**
     * Removes the turns contained in the current prohibition set from the specified
     * <code>server_graph</code>, deleting servers that are no longer connected.
     * Updates the corresponding mappings between the network graph and the server_graph.
     * @param server_graph the server graph in which the prohibited turns shall be removed
     * @param link_server_map the mapping from links in the network graph to servers in the
     * server graph; will be modified accordingly
     * @param router_turns_map the mapping from routers in the network graph to servers in the server
     * graph; will be modified accordingly
     */
    public void removeProhibitedTurns(DirectedSparseGraph server_graph,
    		Map link_server_map, Map router_turns_map) {
    	for (Iterator iter = prohibited_turns.iterator(); iter.hasNext(); ) {
    		Pair p = (Pair) iter.next();
    		Vertex server1 = (Vertex) link_server_map.get(p.getFirst());
    		Vertex server2 = (Vertex) link_server_map.get(p.getSecond());
    		if (server1 != null && server2 != null) {
        		DirectedEdge turn  = (DirectedEdge) server1.findEdge(server2);
        		Set turns = (Set) router_turns_map.get(
        				((DirectedEdge) p.getFirst()).getDest());
        		turns.remove(turn);
        		if (turns.size() == 0) {
        			router_turns_map.remove(((DirectedEdge) p.getFirst()).getDest());
        		}
        		server_graph.removeEdge(turn);
        		
        		if (server1 != null && server1.degree() == 0) {
        			for (Iterator del = link_server_map.keySet().iterator(); del.hasNext(); ) {
        				DirectedEdge key = (DirectedEdge) del.next();
        				if (link_server_map.get(key) == server1) {
        					link_server_map.remove(key);
        					break;
        				}
        			}
        			server_graph.removeVertex(server1);
        		}
        		if (server2 != null && server2.degree() == 0) {
        			for (Iterator del = link_server_map.keySet().iterator(); del.hasNext(); ) {
        				DirectedEdge key = (DirectedEdge) del.next();
        				if (link_server_map.get(key) == server2) {
        					link_server_map.remove(key);
        					break;
        				}
        			}
        			server_graph.removeVertex(server2);
        		}
    		}
    	}
    }
    
    /**
     * Adds the turns contained in the current permission set to the specified
     * <code>server_graph</code>, creating servers if necessary. Updates the
     * corresponding mappings between the network graph and the server_graph.
     * @param server_graph the server graph in which the permitted turns will be added
     * @param link_server_map the mapping from links in the network graph to servers in the
     * server graph; will be modified accordingly
     * @param router_turns_map the mapping from routers in the network graph to servers in the server
     * graph; will be modified accordingly
     */
    public void addPermittedTurns(DirectedSparseGraph server_graph,
    		Map link_server_map, Map router_turns_map) {
    	for (Iterator iter = permitted_turns.iterator(); iter.hasNext(); ) {
    		Pair p = (Pair) iter.next();
    		DirectedEdge link1 = (DirectedEdge) p.getFirst();
    		DirectedEdge link2 = (DirectedEdge) p.getSecond();
    		if (!link_server_map.containsKey(link1)) {
    			Vertex v = new SimpleSparseVertex();
    			link_server_map.put(link1, v);
    			server_graph.addVertex(v);
    		}
    		if (!link_server_map.containsKey(link2)) {
    			Vertex v = new SimpleSparseVertex();
    			link_server_map.put(link2, v);
    			server_graph.addVertex(v);
    		}
    		Vertex server1 = (Vertex) link_server_map.get(p.getFirst());
    		Vertex server2 = (Vertex) link_server_map.get(p.getSecond());
    		DirectedEdge turn  = (DirectedEdge) server1.findEdge(server2);
    		if (turn == null) {
				turn = new DirectedSparseEdge(server1, server2);
				if (!router_turns_map.containsKey(link1.getDest())) {
					router_turns_map.put(link1.getDest(), new HashSet());
				}
				((Set) router_turns_map.get(link1.getDest())).add(turn);
				server_graph.addEdge(turn);
    		}
    	}
    }
    
    /**
     * Runs a single iteration of the TP algorithm for the specified component.
     * @param remaining_nodes
     * @param special_node
     */
    private void runTurnProhibitionInternal(Set remaining_nodes, Vertex special_node) {
    	Vertex selected_node = selectNodeOfMinimalDegree( remaining_nodes, special_node );

    	prohibitTurnsAround(selected_node);

       	permitTurnsStarting(selected_node);

       	Set components = getComponentsOfConnectivity(remaining_nodes, selected_node);

       	Map special_links = null;
       	if (components.size() > 1) {
       		special_links = chooseSpecialLinks(selected_node, components);
       		allowTurnsBetweenSpecialLinks(selected_node, special_links);
       	}
       	
       	visited_nodes.add(selected_node);
       	visited_links.addAll(selected_node.getIncidentEdges());
       	removeNode(selected_node);
       	
       	if (remaining_nodes.size() <= 2) {
           	remaining_nodes = SetUtils.getDifference(remaining_nodes, Collections.singleton(selected_node));
       		if (!remaining_nodes.isEmpty()) {
           		visited_nodes.add(remaining_nodes);
       			removeNode((Vertex) remaining_nodes.iterator().next());
       		}
       		return;
       	}
       	
       	for (Iterator iter = components.iterator(); iter.hasNext(); ) {
       		Set component = (Set) iter.next();
       		Vertex next_special_node = special_node;
       		if (components.size() > 1 && !component.contains(special_node)) {
       			next_special_node = ((DirectedEdge) special_links.get(component)).getDest();
       		}
  			runTurnProhibitionInternal(component, next_special_node);
       	}
    }

    /**
     * Removes the specified node from the node queue and node-to-node-degree-map.
     * @param v
     */
    private void removeNode(Vertex v) {
    	for (Iterator iter = v.getNeighbors().iterator(); iter.hasNext(); ) {
    		Vertex neighbor = (Vertex) iter.next();
    		if (node_queue.contains(neighbor)) {
    			int old_degree = ((Integer) node_degrees.get(neighbor)).intValue();
    			node_degrees.put(neighbor, new Integer(old_degree-1));
    		}
    	}
		node_degrees.remove(v);
		node_queue.remove(v);
    	Collections.sort(node_queue, node_comparator);
    }
    
    /**
     * Determines the components of connectivity after removing a node. Returns
     * a set of components, each component consisting of a set of nodes.
     * @param nodes the remaining nodes in the graph
     * @param v the node that will be removed
     * @return a set of components
     */
    private Set getComponentsOfConnectivity(Set nodes, Vertex v) {
    	Set remaining_nodes     = new HashSet(nodes);
    	Set remaining_neighbors = SetUtils.getIntersection(v.getNeighbors(), remaining_nodes);

    	Set components = new HashSet();
    	while (remaining_neighbors.size() > 0) {
    		Vertex source = (Vertex) remaining_neighbors.iterator().next();
    		Set connected_nodes = doBFS(source, v);
    		components.add(connected_nodes);
    		remaining_neighbors.removeAll(connected_nodes);
    	}
    	return components;
    }
    
    /**
     * Performs a breadth-first-search starting from the specified source. Returns a
     * set of all nodes that have not yet been visited by the TP algorithm,
     * are not equal to <code>hidden</code>, and that are connected to the source node.
     * @param source the root of the BFS
     * @param hidden
     * @return a set of connected nodes
     */
    private Set doBFS(Vertex source, Vertex hidden) {
    	Set connected_nodes = new HashSet();
        List current_list = new ArrayList();
        connected_nodes.add(source);
        current_list.add(source);
        while (true) {
        	List next_list = new ArrayList();
        	for (Iterator iter = current_list.iterator(); iter.hasNext(); ) {
                Vertex node = (Vertex) iter.next();
                for (Iterator iter2 = node.getSuccessors().iterator(); iter2.hasNext();) {
                	Vertex neighbor = (Vertex) iter2.next();
                    if (!connected_nodes.contains(neighbor)
                    		&& !visited_nodes.contains(neighbor)
                    		&& neighbor != hidden) {
                    	connected_nodes.add(neighbor);
                    	next_list.add(neighbor);
                    }
                }
            }
            if (next_list.size() == 0) {
            	break;
            }
            current_list = next_list;
        }
    	return connected_nodes;
    }
    
    /**
     * Selects a special link between <code>v</code> and each component of connectivity
     * and returns a component-to-special-link map. The special link is always one
     * whose source vertex is <code>v</code> and whose destination vertex is in the
     * component of connectivity.  
     * @param v
     * @param components
     * @return a mapping from components to special links
     */
    private Map chooseSpecialLinks(Vertex v, Set components) {
    	Map special_links = new HashMap();
    	for (Iterator iter = components.iterator(); iter.hasNext(); ) {
    		Set component = (Set) iter.next();
    		Set neighbors = SetUtils.getIntersection(component, v.getNeighbors());
    		Vertex neighbor = (Vertex) neighbors.iterator().next();
    		special_links.put(component, v.findEdge(neighbor));
    	}
    	return special_links;
    }
    
    /**
     * Allows all turns between special links.
     * @param v
     * @param special_links
     */
    private void allowTurnsBetweenSpecialLinks(Vertex v, Map special_links) {
    	Collection links = special_links.values();
    	for (Iterator iter1 = links.iterator(); iter1.hasNext(); ) {
    		DirectedEdge e  = (DirectedEdge) iter1.next();
    		DirectedEdge in = (DirectedEdge) e.getDest().findEdge(v);
    		for (Iterator iter2 = links.iterator(); iter2.hasNext(); ) {
    			DirectedEdge out = (DirectedEdge) iter2.next();
    			if (in.getSource() != out.getDest()) {
    				permitted_turns.add(new Pair(in, out));
    				prohibited_turns.remove(new Pair(in, out));
    			}
    		}
    	}
    }
    
    /**
     * Selects a node of minimal degree from the set of remaining (= unvisited)
     * nodes in the given component, excluding the special node. Prefers nodes
     * that do not have special nodes as neighbors.
     * @param component the nodes in the current component
     * @param special the current special node
     * @return the selected node
     */
    private Vertex selectNodeOfMinimalDegree(Set component, Vertex special) {
    	Vertex selected_node = null;
    	for (Iterator iter = node_queue.iterator(); iter.hasNext(); ) {
    		Vertex node = (Vertex) iter.next();
    		if (!component.contains(node) || node == special) {
    			continue;
    		}
    		if (selected_node != null
    				&& ((Integer) node_degrees.get(node)).intValue()
    				    > ((Integer) node_degrees.get(selected_node)).intValue()) {
    			break;
    		}
    		selected_node = node;
    		if (!hasSpecialNeighbor(selected_node, special)) {
    			break;
    		}
    	}
    	return selected_node;
    }
    
    /**
     * Adds all turns around vertex <code>v</code> to the prohibition set.
     * @param v
     */
    private void prohibitTurnsAround(Vertex v) {
    	Set inedges  = SetUtils.getDifference(v.getInEdges(), visited_links);
    	Set outedges = SetUtils.getDifference(v.getOutEdges(), visited_links);
    	
    	for (Iterator inIter = inedges.iterator(); inIter.hasNext(); ) {
    		DirectedEdge in = (DirectedEdge) inIter.next();
    		
        	for (Iterator outIter = outedges.iterator(); outIter.hasNext(); ) {
        		DirectedEdge out = (DirectedEdge) outIter.next();
        		
        		if (in.getSource() != out.getDest()) {
               		prohibited_turns.add(new Pair(in, out));        			
				}
        	}
    	}
    }
    
    /*
     * Adds all turns starting from vertex <code>v</code> to the permission set.
     * @param v
     */
    private void permitTurnsStarting(Vertex v) {
    	Set outedges = SetUtils.getDifference(v.getOutEdges(), visited_links);    	
    	for (Iterator iter = outedges.iterator(); iter.hasNext(); ) {
    		DirectedEdge out = (DirectedEdge) iter.next();
    		
        	Set outedges2 = SetUtils.getDifference(out.getDest().getOutEdges(), visited_links);    	
        	for (Iterator iter2 = outedges2.iterator(); iter2.hasNext(); ) {
        		DirectedEdge out2 = (DirectedEdge) iter2.next();

        		if (out2.getDest() != v) {
        			DirectedEdge out_rev  = reversed(out);
        			DirectedEdge out2_rev = reversed(out2);
        			if (out_rev == null || out2_rev == null) {
        				throw new RuntimeException("Need a bidirected graph to run turn-prohibition!");
        			}
               		permitted_turns.add(new Pair(out, out2));
               		permitted_turns.add(new Pair(out2_rev, out_rev));
				}
        	}
    	}    	
    }

    /**
     * Tests whether the <code>special</code> node is a neighbor of <code>v</code>.
     * @param v
     * @param special
     * @return <code>true</code> if <code>special</code> is <code>v</code>'s neighbor,
     * <code>false</code> otherwise
     */
    private boolean hasSpecialNeighbor(Vertex v, Vertex special) {
    	return SetUtils.getIntersection( v.getNeighbors(), Collections.singleton(special) ).size() > 0;
    }

    /**
     * Returns the reversed edge, i.e. if <code>e</code> is an edge (u,v) returns (v,u).
     * @param e the edge to be reversed
     * @return the reverse edge
     */
    private DirectedEdge reversed(DirectedEdge e) {
    	return (DirectedEdge) e.getDest().findEdge(e.getSource());
    }
}
